import os.path

import objc
from AppKit import NSWorkspace  # pylint: disable=no-name-in-module
from Foundation import NSURL, NSArray  # pylint: disable=no-name-in-module
from UniformTypeIdentifiers import (  # pylint: disable=no-name-in-module
    UTTagClassFilenameExtension,
    UTType,
)


class ExtHasNoRegisteredUTI(ValueError):
    """
    Raised when trying to set a handler for a file extension
    that does not have an associated registered UTI.
    """


class ApplicationNotFound(ValueError):
    """
    Raised when a handler reference cannot be resolved.
    """


class BundleURLNotFound(ApplicationNotFound):
    """
    Raised when a bundle ID is not registered on the system.
    """


class Dooti:
    """
    Wrapper for macOS system API to manage default handlers on macOS 12.0+.
    """

    def __init__(self, workspace=None):
        if workspace is None:
            workspace = NSWorkspace.sharedWorkspace()

        self.workspace = workspace

    @staticmethod
    def ext_to_utis(ext: str) -> NSArray:
        """
        Returns all UTI associated with specified file extension.
        If the extension is not registered with MacOS, will return
        a dynamic UTI as first and only element.

        :param str ext: file extension to look up associated UTI for
        """
        return UTType.typesWithTag_tagClass_conformingToType_(
            ext, UTTagClassFilenameExtension, objc.nil
        )

    def set_default_uti(self, uti: str | UTType, app: str) -> None:
        """
        Sets a default handler for a specific UTI.

        :param str | UTType uti: UTI to set the default handler for
        :param str app: absolute filesystem path, name or bundle ID of the handler
        """
        if not isinstance(uti, UTType):
            uti = UTType.importedTypeWithIdentifier_(uti)

        path = self.get_app_path(app)

        self.workspace.setDefaultApplicationAtURL_toOpenContentType_completionHandler_(
            path, uti, objc.nil
        )

    def set_default_scheme(self, scheme: str, app: str) -> None:
        """
        Sets a default handler for a specific URL scheme.

        :param str scheme: URL scheme to set the default handler for
        :param str app: absolute filesystem path, name or bundle ID of the handler
        """

        path = self.get_app_path(app)

        self.workspace.setDefaultApplicationAtURL_toOpenURLsWithScheme_completionHandler_(
            path, scheme, objc.nil
        )

    def set_default_ext(self, ext: str, app: str, allow_dynamic: bool = False) -> None:
        """
        Sets a default handler for all UTI registered to a file extension.

        :param str ext: file extension to set the default handler for
        :param str app: absolute filesystem path, name or bundle ID of the handler
        :param bool allow_dynamic: whether to allow dynamic UTIs (default False)

        :raises:
            ExtHasNoRegisteredUTI if the file extension is unknown to MacOS and not allowing dynamic UTI
        """
        utis = Dooti.ext_to_utis(ext)

        if self.is_dynamic_uti(utis[0]) and not allow_dynamic:
            raise ExtHasNoRegisteredUTI(
                f"No UTI are registered for file extension '{ext}'. "
                "To force using a dynamic UTI, pass allow_dynamic=True."
            )

        for uti in utis:
            self.set_default_uti(uti, app)

    def is_dynamic_uti(self, ext_or_uti: str | UTType) -> bool:
        """
        Checks whether a UTI is dynamic/whether a file extension is not
        associated with at least one registered UTI.

        :param str | UTType ext_or_uti: UTI or file extension to check
        """
        if isinstance(ext_or_uti, str):
            ext_or_uti = Dooti.ext_to_utis(ext_or_uti)[0]
        return str(ext_or_uti).startswith("dyn.")

    def get_default_uti(self, uti: str | UTType) -> str | None:
        """
        Returns the filesystem path to the default handler for the
        specified UTI.

        :param str | UTType uti: UTI to look up the default handler path for
        """

        if not isinstance(uti, UTType):
            uti = UTType.importedTypeWithIdentifier_(uti)

        handler = self.workspace.URLForApplicationToOpenContentType_(uti)

        if not handler:
            return None

        # name = handler.lastPathComponent()[:-4]

        return handler.fileSystemRepresentation().decode()

    def get_default_ext(self, ext: str) -> str | None:
        """
        Returns the filesystem path to the default handler for the
        specified file extension.

        :param str ext: filename extension to look up the default handler path for
        """
        utis = Dooti.ext_to_utis(ext)

        # assume the handler is the same for all types (sensible?)
        # even if the extension was not registered, utis will still contain
        # a dynamic UTI, so we do not need to check for an empty iterator
        handler = self.workspace.URLForApplicationToOpenContentType_(utis[0])

        if not handler:
            return None

        # name = handler.lastPathComponent()[:-4]

        return handler.fileSystemRepresentation().decode()

    def get_default_scheme(self, scheme: str) -> str | None:
        """
        Returns the filesystem path to the default handler for the
        specified URL scheme.

        :param str ext: filename extension to look up the default handler path for
        """
        if "file" == scheme:
            raise ValueError("The file:// scheme cannot be looked up.")

        url = NSURL.URLWithString_(scheme + "://nonexistant")

        handler = self.workspace.URLForApplicationToOpenURL_(url)

        if not handler:
            return None

        # name = handler.lastPathComponent()[:-4]

        return handler.fileSystemRepresentation().decode()

    def get_app_path(self, app: str) -> NSURL:
        """
        Returns a URL (filesystem path prefixed with 'file://' scheme) to an
        application specified by name, absolute path or bundle ID.

        :param str app: name, absolute filesystem path or bundle ID to look up the URL for

        :raises:
            ApplicationNotFound: when no matching application was found
        """
        if "/" == app[0]:
            return NSURL.fileURLWithPath_(app)

        try:
            return self.bundle_to_url(app)
        except BundleURLNotFound:
            pass

        try:
            return self.name_to_url(app)
        except ApplicationNotFound as exc:
            raise ApplicationNotFound(
                f"Could not find an application matching the description '{app}'."
            ) from exc

    def bundle_to_url(self, bundle_id: str) -> NSURL:
        """
        Returns a URL (filesystem path prefixed with 'file://' scheme) to an
        application with the specified bundle ID.

        :param str bundle_id: bundle ID to look up the URL for

        :raises:
            BundleURLNotFound: when no application with specified bundle ID was found
        """
        path = self.workspace.URLForApplicationWithBundleIdentifier_(bundle_id)

        if path is None:
            raise BundleURLNotFound(
                f"There is no bundle with the identifier '{bundle_id}'."
            )

        return path

    def name_to_url(self, app_name: str) -> NSURL:
        """
        Returns a URL (filesystem path prefixed with 'file://' scheme) to an
        application with the specified name.

        :param str app_name: application name to look up the URL for

        :raises:
            ApplicationNotFound: when no application with specified bundle ID was found
        """
        path = self.workspace.fullPathForApplication_(app_name)

        if path is None:
            raise ApplicationNotFound(
                f"Could not find an application named '{app_name}'."
            )

        return NSURL.fileURLWithPath_(path)

    def path_to_url(self, path: str, skip_check: bool = False) -> NSURL:
        """
        Translates an absolute filesystem path to a URL.

        :param str path: absolute filesystem path to translate
        :param bool skip_check: override check if the target exists and is a directory

        :raises:
            ApplicationNotFound: when no application with specified bundle ID was found
        """

        if not skip_check and not os.path.isdir(path):
            raise ApplicationNotFound(f"Could not find an application in '{path}'.")

        return NSURL.fileURLWithPath_(path)



__author__ = "jeanluc"
__version__ = "0.2.1"

from .dooti import ApplicationNotFound, BundleURLNotFound, Dooti, ExtHasNoRegisteredUTI

__all__ = ["ApplicationNotFound", "BundleURLNotFound", "Dooti", "ExtHasNoRegisteredUTI"]



import argparse
import json
import logging
import sys
import time
from pathlib import Path

import yaml
from xdg import xdg_config_home

from .dooti import ApplicationNotFound, Dooti

log = logging.getLogger(__name__)
logging.basicConfig(
    stream=sys.stderr, level=logging.INFO, format="{levelname}: {message}", style="{"
)


class DootiCLI:
    """
    Wraps Dooti for the command line.
    """

    changes = {}
    errors = []
    handlers = {}
    scopes = ("ext", "scheme", "uti")

    def __init__(self, assume_yes=False, dry_run=False, fmt="yaml"):
        self.do = Dooti()
        self.assume_yes = assume_yes
        self.dry_run = dry_run
        self.fmt = fmt

    def apply_(self, file=None, dynamic=False):
        """
        Apply configuration from a file.
        """
        file = self._find_config(file)
        definitions = self._load_config(file)

        parsed = {"extensions": {}, "schemes": {}, "utis": {}}

        for ext, handler in definitions.get("ext", {}).items():
            try:
                _, single = self.ext([ext], dynamic=dynamic, handler=handler)
                parsed["extensions"].update(single["extensions"])
            except ApplicationNotFound as err:
                self.errors.append(str(err))

        for scope in ("scheme", "uti"):
            parser = getattr(self, scope)
            for item, handler in definitions.get(scope, {}).items():
                try:
                    _, single = parser([item], handler=handler)
                    parsed[f"{scope}s"].update(single[f"{scope}s"])
                except ApplicationNotFound as err:
                    self.errors.append(str(err))

        for handler, app_config in definitions.get("app", {}).items():
            if "ext" in app_config:
                try:
                    _, single = self.ext(
                        app_config["ext"], dynamic=dynamic, handler=handler
                    )
                    parsed["extensions"].update(single["extensions"])
                except ApplicationNotFound as err:
                    self.errors.append(str(err))
            for scope in ("scheme", "uti"):
                if scope in app_config:
                    parser = getattr(self, scope)
                    try:
                        _, single = parser(app_config[scope], handler=handler)
                        parsed[f"{scope}s"].update(single[f"{scope}s"])
                    except ApplicationNotFound as err:
                        self.errors.append(str(err))

        return None, parsed

    def ext(self, extensions, dynamic=False, handler=None):
        """
        Set handler or get handlers for a list of file extensions.
        """
        current = {ext: self.do.get_default_ext(ext) for ext in extensions}
        if handler is None:
            return current, None

        handler = self._lookup_handler(handler)
        if not dynamic:
            allowed_extensions = [
                ext for ext in extensions if not self.do.is_dynamic_uti(ext)
            ]
            disallowed_extensions = set(extensions) - set(allowed_extensions)
            for ext in disallowed_extensions:
                self.errors.append(
                    f"No UTI are registered for file extension '{ext}'. "
                    "To force using a dynamic UTI, pass `-u`/`--dynamic`."
                )
            extensions = allowed_extensions

        diff = {
            ext: {"from": current[ext], "to": handler}
            for ext in extensions
            if current[ext] != handler
        }

        return current, {"extensions": diff}

    def scheme(self, schemes, handler=None):
        """
        Set handler or get handlers for a list of schemes.
        """
        current = {scheme: self.do.get_default_scheme(scheme) for scheme in schemes}

        if handler is None:
            return current, None

        handler = self._lookup_handler(handler)

        diff = {
            scheme: {"from": current[scheme], "to": handler}
            for scheme in schemes
            if current[scheme] != handler
        }

        return current, {"schemes": diff}

    def uti(self, utis, handler=None):
        """
        Set handler or get handlers for a list of UTI.
        """
        current = {uti: self.do.get_default_uti(uti) for uti in utis}

        if handler is None:
            return current, None

        handler = self._lookup_handler(handler)

        diff = {
            uti: {"from": current[uti], "to": handler}
            for uti in utis
            if current[uti] != handler
        }

        return current, {"utis": diff}

    def run(self, func, args):
        """
        Call the requested function, catch errors and handle output.
        """
        ret = None
        try:
            current, diff = getattr(self, func)(**vars(args))
            if diff is None:
                ret = current
            else:
                self._apply_diff(diff)
        except (ValueError, yaml.parser.ParserError, ApplicationNotFound) as err:
            self.errors.append(str(err))
        except Exception as err:  # pylint: disable=broad-except
            log.error(str(err))
        finally:
            self._output(ret)
            # bandaid for PyThread_exit_thread / pthread_exit being called too early
            # because pyobjc does not have the correct metadata for completionHandler
            if ret is None:
                time.sleep(0.1)
            sys.exit(int(bool(self.errors)))

    def _apply_diff(self, diff):
        if self.dry_run or not any(
            (scope in diff and diff[scope])
            for scope in ("extensions", "schemes", "utis")
        ):
            self.changes = diff
            return
        if not (self.assume_yes or self._ask_consent(diff)):
            log.info("Did not get consent to apply changes. Exiting.")
            return

        if "extensions" in diff:
            for ext, handler in diff["extensions"].items():
                self.do.set_default_ext(ext, handler["to"], allow_dynamic=True)
            self.changes["extensions"] = diff["extensions"]

        if "schemes" in diff:
            for scheme, handler in diff["schemes"].items():
                self.do.set_default_scheme(scheme, handler["to"])
            self.changes["schemes"] = diff["schemes"]

        if "utis" in diff:
            for uti, handler in diff["utis"].items():
                self.do.set_default_uti(uti, handler["to"])
            self.changes["utis"] = diff["utis"]

    def _output(self, ret=None):
        if ret is None:
            ret = {"changes": self.changes, "errors": self.errors}
        if "json" == self.fmt:
            return print(json.dumps(ret))
        return print(yaml.dump(ret))

    def _lookup_handler(self, handler):
        if handler not in self.handlers:
            self.handlers[handler] = (
                self.do.get_app_path(handler).fileSystemRepresentation().decode()
            )
        return self.handlers[handler]

    def _find_config(self, file=None):
        if file is None:
            xch = xdg_config_home()

            for path in (
                xch / "dooti.yaml",
                xch / "dooti.yml",
                xch / "dooti" / "dooti.yaml",
                xch / "dooti" / "dooti.yml",
                xch / "dooti" / "config.yaml",
                xch / "dooti" / "config.yml",
            ):
                if path.exists():
                    return path
            raise ValueError(f"Could not find dooti configuration in `{xch}`.")
        if not Path(file).exists():
            raise ValueError(
                f"Passed dooti configuration file `{file}` does not exist."
            )
        return file

    def _load_config(self, file):
        with open(file, encoding="utf-8") as f:
            definitions = yaml.load(f, Loader=yaml.Loader)

        if not isinstance(definitions, dict):
            raise ValueError("Invalid configuration, must be a dictionary.")
        if not any(x in definitions for x in ("app",) + self.scopes):
            raise ValueError(
                "Configuration does not contain any actionable definitions."
            )
        return definitions

    def _ask_consent(self, diffs):
        for atype, diff in diffs.items():
            if not diff:
                continue
            print(
                f"The following {atype[:-1]}{'s are' if len(diff) > 1 else ' is'} "
                "set to be changed:"
            )
            for name, change in diff.items():
                print(f"{name}: {change['from']} -> {change['to']}")
            print()
        try:
            return input("Do you want to continue? [y/n]: ").lower().startswith("y")
        except KeyboardInterrupt:
            print()
            return False


def main():
    """
    Prepare CLI args parser and hand off to DootiCLI
    """
    parser = argparse.ArgumentParser(
        prog="dooti", description="Manage default handlers on macOS."
    )
    parser.add_argument(
        "-f",
        "--format",
        help="The output format. Defaults to YAML.",
        dest="fmt",
        choices=("json", "yaml"),
    )
    parser.add_argument(
        "-y",
        "--yes",
        help="Do not ask for consent, assume yes.",
        dest="assume_yes",
        action="store_true",
    )
    parser.add_argument(
        "-t",
        "--dry-run",
        help="Only show planned changes and exit.",
        dest="dry_run",
        action="store_true",
    )
    subparsers = parser.add_subparsers(help="commands")

    apply_parser = subparsers.add_parser(
        "apply",
        help="Apply a YAML state configuration.",
    )
    apply_parser.add_argument(
        "-u",
        "--dynamic",
        action="store_true",
        help="Allow unregistered file extensions / dynamic UTIs.",
    )
    apply_parser.add_argument(
        "-i",
        "--file",
        help="Configuration to apply. If unspecified, searches in $XDG_CONFIG_HOME.",
    )
    apply_parser.set_defaults(func="apply_")

    ext_parser = subparsers.add_parser(
        "ext",
        help="Manage the default handler for all UTI associated with file extensions",
    )
    ext_parser.add_argument(
        "extensions", nargs="+", help="File extension(s) to operate on"
    )
    ext_parser.add_argument(
        "-u",
        "--dynamic",
        action="store_true",
        help="Allow unregistered file extensions / dynamic UTIs.",
    )
    ext_parser.add_argument(
        "-x",
        "--handler",
        help="The handler to associate with the file extensions. If unset, returns the default handler(s).",
    )
    ext_parser.set_defaults(func="ext")

    scheme_parser = subparsers.add_parser(
        "scheme", help="Manage default handler for URI scheme(s)"
    )
    scheme_parser.add_argument("schemes", nargs="+", help="Scheme(s) to operate on")
    scheme_parser.add_argument(
        "-x",
        "--handler",
        help="The handler to associate with the scheme(s). If unset, returns the default handler(s).",
    )
    scheme_parser.set_defaults(func="scheme")

    uti_parser = subparsers.add_parser("uti", help="Manage default handler for UTI(s)")
    uti_parser.add_argument("utis", nargs="+", help="UTI(s) to operate on")
    uti_parser.add_argument(
        "-x",
        "--handler",
        help="The handler to associate with the UTI(s). If unset, returns the default handler(s).",
    )
    uti_parser.set_defaults(func="uti")

    args = parser.parse_args()
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()
    args = parser.parse_args()
    cli = DootiCLI(assume_yes=args.assume_yes, dry_run=args.dry_run, fmt=args.fmt)
    func = args.func
    del args.func
    del args.assume_yes
    del args.dry_run
    del args.fmt

    cli.run(func, args)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dooti']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'pyobjc-core>=7.2,<8.0',
 'pyobjc-framework-UniformTypeIdentifiers>=7.2,<8.0',
 'xdg>=5.1.1,<6.0.0']

entry_points = \
{'console_scripts': ['dooti = dooti.cli:main']}

setup_kwargs = {
    'name': 'dooti',
    'version': '0.2.1',
    'description': 'Manage default file and URI scheme handlers on macOS 12.0+',
    'long_description': '==============\nDooti overview\n==============\n\nManage default handlers for files and URI schemes on MacOS 12.0+.\n\n\nFeatures\n--------\n* Manage handlers by specifying file extension, UTI or URI scheme.\n* Specify handlers by name, bundle ID or absolute path.\n* Check your associations into a dotfiles repository and ensure an idempotent state by simply running ``dooti apply``.\n\n\nInstallation\n------------\nI recommend installing with `pipx <https://pypa.github.io/pipx/>`_, although pip will work fine as well:\n\n.. code-block:: bash\n\n        pipx install dooti\n\n\nQuickstart\n----------\n``dooti`` currently supports three specific subcommands (``ext``, ``scheme``, ``uti``) and a broad one (``apply``).\n\nLookup handlers\n~~~~~~~~~~~~~~~\nSimply pass a list of one specific type to the appropriate command. Example:\n\n.. code-block:: console\n\n    $ dooti ext html xml\n    html: /Applications/Firefox.app\n    xml: /Applications/Sublime Text.app\n\n    $ dooti --format json scheme http https ftp | jq\n    {\n      "http": "/Applications/Firefox.app",\n      "https": "/Applications/Firefox.app",\n      "ftp": "/System/Library/CoreServices/Finder.app"\n    }\n\n    $ dooti uti public.html\n    public.html: /Applications/Firefox.app\n\nSet handlers\n~~~~~~~~~~~~\nChanging the default handler can be requested by adding ``-x <handler_ref>`` to the lookup command. The handler reference can be a name, bundle ID or absolute filesystem path. Example:\n\n.. code-block:: console\n\n    $ dooti ext csv xml -x "Sublime Text"\n    The following extensions are set to be changed:\n    csv: /Applications/Numbers.app -> /Applications/Sublime Text.app\n    xml: /Applications/Firefox.app -> /Applications/Sublime Text.app\n\n    $ dooti scheme ftp -x /Applications/Firefox.app\n    The following scheme is set to be changed:\n    ftp: /System/Library/CoreServices/Finder.app -> /Applications/Firefox.app\n\n    $ dooti -tf json uti public.c-source -x com.sublimetext.4\n    {"changes": {"utis": {"public.c-source": {"from": "/Applications/Notes.app", "to": "/Applications/Sublime Text.app"}}}, "errors": []}\n\n\nEnsure state\n~~~~~~~~~~~~\n``dooti`` can ingest and apply a YAML configuration like this:\n\n.. code-block:: yaml\n\n    ext:\n      jpeg: Preview\n\n    scheme:\n      http: Firefox\n      mailto: Mail\n\n    uti:\n      public.c‑source: Sublime Text\n\n    app:\n      Sublime Text:\n        ext:\n          - py\n          - rst\n          - yml\n          - yaml\n        uti:\n          - public.fortran‑source\n\n      Brave Browser:\n        scheme:\n          - ipfs\n\nBy default, it looks at ``$XDG_CONFIG_HOME/dooti/config.yaml`` (and others, see ``docs/usage.rst``).\n\nLimitations\n-----------\n* The designated handler has to be installed before running the command.\n* Setting some URI scheme handlers (especially for http) might cause a prompt.\n* Setting some file extension handlers might be restricted (especially html seems to fail silently).\n\n\nWhy?\n----\nMost existing tools use `LSSetDefaultRoleHandlerForContentType <https://developer.apple.com/documentation/coreservices/1444955-lssetdefaultrolehandlerforconten>`_ and `LSSetDefaultHandlerForURLScheme <https://developer.apple.com/documentation/coreservices/1447760-lssetdefaulthandlerforurlscheme?language=objc>`_, which are deprecated in macOS 12.0. ``dooti`` uses a more recent API and should work on Monterey (12.0) and above.\n\n\nSimilar tools\n-------------\n* `duti <https://github.com/moretension/duti>`_\n* `openwith <https://github.com/jdek/openwith>`_\n* `defaultbrowser <https://gist.github.com/miketaylr/5969656>`_\n* `SwiftDefaultApps <https://github.com/Lord-Kamina/SwiftDefaultApps>`_\n',
    'author': 'jeanluc',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/lkubb/dooti',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
